# cython:language_level = 3
# distutils: language = c++
import cython

from libc.stdint cimport uint64_t, uint32_t, uint16_t, uint8_t
from libcpp.vector cimport vector
from libcpp cimport bool

import operator
from functools import reduce
import sys

cimport numpy as cnp
import numpy as np
from zmesh.mesh import Mesh

cdef extern from "cMesher.hpp":
  cdef struct MeshObject:
    vector[float] points
    vector[float] normals
    vector[unsigned int] faces

  cdef cppclass CMesher[P,L,S]:
    CMesher(vector[float] voxel_res) except +
    void mesh(
      L* data, 
      unsigned int sx, 
      unsigned int sy, 
      unsigned int sz, 
      bool c_order
    )
    vector[L] ids()
    MeshObject get_mesh(
      L segid, 
      bool normals,
      int simplification_factor, 
      float max_simplification_error, 
      float min_simplification_error,
      bool transpose
    )
    # NOTE: need to define triangle_t
    MeshObject simplify_points(
      uint64_t* points, 
      size_t Nv, 
      bool normals, 
      int simplification_factor, 
      float max_simplification_error,
      float min_simplification_error,
    )
    bool erase(L segid)
    void clear()
    P pack_coords(P x, P y, P z)

class Mesher:
  """
  Represents a meshed volume. 

  Call mesher.mesh(labels) then you can 
  extract meshes from it using mesher.get(label).
  """
  def __init__(self, voxel_res):
    self.voxel_res = voxel_res
    self._mesher = Mesher6464(self.voxel_res)

  @property
  def voxel_res(self):
    return self._voxel_res

  @voxel_res.setter
  def voxel_res(self, res):
    self._voxel_res = np.array(res, dtype=np.float32)

  @cython.binding(True)
  def mesh(self, data, close=False):
    """
    Triggers the multi-label meshing process.
    After the mesher has run, you can call mesher.get.

    data: 3d numpy array
    close: By default, meshes flush against the edge of
      the image will be left open on the sides that touch
      the boundary. If True, this ensures that all meshes
      produced will be closed.
    """
    del self._mesher

    if not data.flags.c_contiguous and not data.flags.f_contiguous:
      data = np.ascontiguousarray(data)

    if close:
      tmp = np.zeros(np.array(data.shape) + 2, dtype=data.dtype, order="C")
      tmp[1:-1,1:-1,1:-1] = data
      data = tmp
      del tmp

    shape = np.array(data.shape)
    nbytes = np.dtype(data.dtype).itemsize

    # z is allocated 10 position bits, y 11, and x 11 bits.
    # The last bit is 2^-1 (fixed point) so divide by two to
    # get the maximum supported shape. 
    if shape[0] > 1023 or shape[1] > 1023 or shape[2] > 511:
      MesherClass = Mesher6464
      if nbytes == 4:
        MesherClass = Mesher6432
      elif nbytes == 2:
        MesherClass = Mesher6416
      elif nbytes == 1:
        MesherClass = Mesher6408
    else:
      MesherClass = Mesher3264
      if nbytes == 4:
        MesherClass = Mesher3232
      elif nbytes == 2:
        MesherClass = Mesher3216
      elif nbytes == 1:
        MesherClass = Mesher3208

    self._mesher = MesherClass(self.voxel_res)

    return self._mesher.mesh(data)

  @cython.binding(True)
  def ids(self):
    return self._mesher.ids()

  @cython.binding(True)
  def get_mesh(
    self, mesh_id, normals=False, 
    simplification_factor=0, max_simplification_error=40,
    voxel_centered=False
  ):
    """
    NOTE: This function is deprecated. Please use "get" as 
    this function transposes the mesh.

    mesh_id: the integer id of the mesh
    normals: whether to calculate vertex normals for the mesh
    simplification_factor: Try to reduce the number of triangles by this factor.
    max_simplification_error: maximum distance vertices are allowed to deviate from
      their original position. Units are physical, not voxels.
    voxel_centered: By default, the meshes produced will be centered at 0,0,0.
      If enabled, the meshes will be centered in the voxel at 0.5,0.5,0.5.

    Returns: Mesh
    """
    mesh = self._mesher.get_mesh(
      mesh_id, normals, 
      simplification_factor, max_simplification_error, 
      transpose=True
    )

    return self._normalize_simplified_mesh(mesh, voxel_centered, physical=True)

  @cython.binding(True)
  def get(
    self, label, 
    normals = False, 
    reduction_factor = 0, 
    max_error = None,
    voxel_centered = False
  ):
    """
    label: the integer id of the mesh
    normals: whether to calculate vertex normals for the mesh
    reduction_factor: Try to reduce the number of triangles by this integer factor. 
      0 disables simplification.
    max_error: maximum distance vertices are allowed to deviate from
      their original position. Units are physical, not voxels. If None, automatically
      is set to at most one voxel distortion based on the provided resolution.
    voxel_centered: By default, the meshes produced will be centered at 0,0,0.
      If enabled, the meshes will be centered in the voxel at 0.5,0.5,0.5.

    Returns: Mesh
    """
    # automatically select one voxel max distortion
    if max_error is None:
      max_error = self.voxel_res.max()

    mesh = self._mesher.get_mesh(
      label, normals, 
      reduction_factor, max_error,
      transpose=False
    )

    mesh = self._normalize_simplified_mesh(mesh, voxel_centered, physical=True)
    mesh.id = int(label)
    return mesh

  def _normalize_simplified_mesh(self, mesh, voxel_centered, physical):
    points = np.array(mesh['points'], dtype=np.float32)
    Nv = points.size // 3
    Nf = len(mesh['faces']) // 3

    points = points.reshape(Nv, 3)
    if not physical:
      points *= self.voxel_res

    if voxel_centered:
      points += self.voxel_res
    points /= 2.0
    faces = np.array(mesh['faces'], dtype=np.uint32).reshape(Nf, 3)
    
    normals = None
    if mesh['normals']:
      normals = np.array(mesh['normals'], dtype=np.float32).reshape(Nv, 3)

    return Mesh(points, faces, normals)

  @cython.binding(True)
  def compute_normals(self, mesh):
    """
    Returns: Mesh with normals computed
    """
    return self.simplify(mesh, reduction_factor=0, max_error=0, compute_normals=True)

  # adding this causes a cython compilation error as of 3.0.10
  # @cython.binding(True)
  def simplify(
    self, mesh, 
    int reduction_factor = 0, 
    float max_error = 40, 
    compute_normals = False,
    voxel_centered = False, 
    float min_error = -1,
  ):
    """
    Mesh simplify(
      mesh, reduction_factor=0, 
      max_error=40, compute_normals=False, 
      voxel_centered=False, min_error = -1
    )

    Given a mesh object (either zmesh.Mesh or another object that has
    mesh.vertices and mesh.faces implemented as numpy arrays), apply
    the quadratic edge collapse algorithm. 

    Note: the maximum range spread between vertex values (in x, y, or z)
    is 2^20. The simplifier cannot represent values outside that range.

    Optional:
      reduction_factor: Triangle reduction factor target. If all vertices
        are maxxed out in terms of their error tolerance the algorithm will
        stop short of this target.
      max_error: The maximum allowed displacement of a vertex in physical
        distance.
      min_error: Continue simplifying until this error target is met OR secondarily 
        the target number of triangles is met. Set this to 0 to force the 
        reduction_factor to take precedence. -1 indicates that the c++ default
        value should be used.

        The c++ default value is std::numeric_limits<Float>::epsilon() * 25

      compute_normals: whether or not to also compute the vertex normals
      voxel_centered: By default, the meshes produced will be centered at 
        0,0,0. If enabled, the meshes will be centered in the voxel at 
        0.5,0.5,0.5.

    Returns: Mesh
    """
    mesher = new CMesher[uint64_t, uint64_t, float]((1,1,1))

    cdef size_t ti = 0
    cdef size_t vi = 0
    cdef uint64_t vert = 0

    cdef cnp.ndarray[float, ndim=3] triangles = mesh.vertices[mesh.faces]
    cdef cnp.ndarray[uint64_t, ndim=2] packed_triangles = np.zeros( 
      (triangles.shape[0], 3), dtype=np.uint64, order='C'
    ) 

    max_error //= np.sqrt(np.sum(self.voxel_res ** 2))

    triangles /= self.voxel_res
    triangles *= 2.0
    cdef float min_vertex = np.min(triangles)
    if min_vertex != 0:
      triangles -= min_vertex

    # uint64 // 3 = 21 bits - 1 bit for representing half voxels
    if np.max(triangles) > 2**20:
      raise ValueError(
        "Vertex positions larger than simplifier representation limit (2^20). "
        f"Did you set the resolution correctly? voxel res: {self.voxel_res}"
      )

    cdef size_t Nv = triangles.shape[0]

    for ti in range(Nv):
      for vi in range(3):
        packed_triangles[ti, vi] = mesher.pack_coords(
          <uint64_t>triangles[ti, vi, 0], <uint64_t>triangles[ti, vi, 1], <uint64_t>triangles[ti, vi, 2]
        )
    del triangles

    cdef MeshObject result = mesher.simplify_points(
      <uint64_t*>&packed_triangles[0,0], Nv, 
      <bool>compute_normals, reduction_factor, 
      max_error, min_error
    )
    del mesher

    cdef int i = 0
    if min_vertex != 0:
      for i in range(len(result.points)):
        result.points[i] += min_vertex

    return self._normalize_simplified_mesh(result, voxel_centered, physical=False)
  
  def clear(self):
    self._mesher.clear()
  
  @cython.binding(True) 
  def erase(self, segid):
    return self._mesher.erase(segid)

def reshape(arr, shape, order=None):
  """
  If the array is contiguous, attempt an in place reshape
  rather than potentially making a copy.

  Required:
    arr: The input numpy array.
    shape: The desired shape (must be the same size as arr)

  Optional: 
    order: 'C', 'F', or None (determine automatically)

  Returns: reshaped array
  """
  if order is None:
    if arr.flags['F_CONTIGUOUS']:
      order = 'F'
    elif arr.flags['C_CONTIGUOUS']:
      order = 'C'
    else:
      return arr.reshape(shape)

  cdef int nbytes = np.dtype(arr.dtype).itemsize

  if order == 'C':
    strides = [ reduce(operator.mul, shape[i:]) * nbytes for i in range(1, len(shape)) ]
    strides += [ nbytes ]
    return np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)
  else:
    strides = [ reduce(operator.mul, shape[:i]) * nbytes for i in range(1, len(shape)) ]
    strides = [ nbytes ] + strides
    return np.lib.stride_tricks.as_strided(arr, shape=shape, strides=strides)


# NOTE: THE USER INTERFACE CLASS IS "Mesher" ABOVE

# The following wrapper classes are generated from a template!!

# The pattern is: Mesher[PositionType bits][LabelType bits]
# 64 bit position type can represent very large arrays
# 32 bit can represent 1023 x 1023 x 511 arrays
cdef class Mesher3208:
  cdef CMesher[uint32_t, uint8_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint32_t, uint8_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint8_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint8)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher3216:
  cdef CMesher[uint32_t, uint16_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint32_t, uint16_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint16_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint16)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher3232:
  cdef CMesher[uint32_t, uint32_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint32_t, uint32_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint32_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint32)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher3264:
  cdef CMesher[uint32_t, uint64_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint32_t, uint64_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint64_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint64)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher6408:
  cdef CMesher[uint64_t, uint8_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint64_t, uint8_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint8_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint8)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher6416:
  cdef CMesher[uint64_t, uint16_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint64_t, uint16_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint16_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint16)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher6432:
  cdef CMesher[uint64_t, uint32_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint64_t, uint32_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint32_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint32)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)

cdef class Mesher6464:
  cdef CMesher[uint64_t, uint64_t, float] *ptr      # hold a C++ instance which we're wrapping

  def __cinit__(self, voxel_res):
    self.ptr = new CMesher[uint64_t, uint64_t, float](voxel_res)

  def __dealloc__(self):
    del self.ptr

  def mesh(self, data):
    cdef cnp.ndarray[uint64_t, ndim=1] flat_data = reshape(data, (data.size,)).view(np.uint64)
    self.ptr.mesh(
      &flat_data[0],
      data.shape[0], data.shape[1], data.shape[2],
      data.flags.c_contiguous
    )

  def ids(self):
    return self.ptr.ids()
  
  def get_mesh(self, mesh_id, normals=False, simplification_factor=0, max_simplification_error=40, min_simplification_error=(25 * sys.float_info.epsilon), transpose=True):
    return self.ptr.get_mesh(mesh_id, normals, simplification_factor, max_simplification_error, min_simplification_error, transpose)
  
  def clear(self):
    self.ptr.clear()

  def erase(self, mesh_id):
    return self.ptr.erase(mesh_id)




# ZMESH
# marching cubes (blank): 0.897s, 149.42 MVx/sec, N=1
# marching cubes (filled): 0.966s, 138.78 MVx/sec, N=1
# marching cubes (connectomics.npy 512^3, uint32): 3.132s, 42.78 MVx/sec, N=1
# marching cubes (random 448^3 uint64): 25.092s, 3.55 MVx/sec, N=1
# SKIMAGE
# marching cubes (blank) NOT HANDLED
# marching cubes (filled) NOT HANDLED
# marching cubes (connectomics.npy 512^3, uint32): 5.359s, 25.00 MVx/sec, N=1
# marching cubes (random 448^3 uint64): 69.927s, 1.27 MVx/sec, N=1

# simplification (connectomics.npy): 434.549s, 0.31 MVx/sec, N=1

import numpy as np

import zmesh
import time
from tqdm import tqdm

def result(label, dt, data, N):
    voxels = data.size
    mvx = voxels // (10 ** 6)
    print(f"{label}: {dt:02.3f}s, {N * mvx / dt:.2f} MVx/sec, N={N}")

def test_zmesh_marching_cubes():
    labels = np.zeros((512,512,512), dtype=np.uint8, order="C")
    mesher = zmesh.Mesher((1,1,1))
    N = 1
    start = time.time()
    for _ in range(N):
        mesher.mesh(labels)
    end = time.time()
    result("marching cubes (blank)", end - start, labels, N=N)

    labels = np.ones((512,512,512), dtype=np.uint8, order="C")
    mesher = zmesh.Mesher((1,1,1))
    N = 1
    start = time.time()
    for _ in range(N):
        mesher.mesh(labels, close=True)
    end = time.time()
    result("marching cubes (filled)", end - start, labels, N=N)

    labels = np.load("./connectomics.npy")
    labels = np.ascontiguousarray(labels)
    mesher = zmesh.Mesher((1,1,1))

    N = 1
    start = time.time()
    for _ in range(N):
        mesher.mesh(labels)
    end = time.time()
    result("marching cubes (connectomics.npy)", end - start, labels, N=N)

    labels = np.random.randint(0,1000, size=(448,448,448), dtype=np.uint32)
    # labels = np.ascontiguousarray(labels)
    mesher = zmesh.Mesher((1,1,1))

    N = 1
    start = time.time()
    for _ in range(N):
        mesher.mesh(labels)
    end = time.time()
    result("marching cubes (random)", end - start, labels, N=N)

def test_scikit_marching_cubes():
    import skimage.measure

    print("marching cubes (blank) NOT HANDLED")
    print("marching cubes (filled) NOT HANDLED")
    
    labels = np.ones((512,512,512))
    labels = np.load("./connectomics.npy")
    labels = np.ascontiguousarray(labels)

    N = 1
    start = time.time()
    for _ in range(N):
        skimage.measure.marching_cubes(labels)
    end = time.time()
    result("marching cubes (connectomics.npy)", end - start, labels, N=N)

    labels = np.random.randint(0,1000, size=(448,448,448), dtype=np.uint32)
    labels = np.ascontiguousarray(labels)

    N = 1
    start = time.time()
    for _ in range(N):
        skimage.measure.marching_cubes(labels)
    end = time.time()
    result("marching cubes (random)", end - start, labels, N=N)

# Ran zmesh simplification and summed the sizes
# of the meshes.
# factor 0 max error 0:  1614121164 bytes (1.0x)
# factor 100 max error 0: 503561448 bytes (3.2x)
# factor 100 max error 1: 350636148 bytes (4.6x)

def test_zmesh_simplification():
    labels = np.load("./connectomics.npy")
    mesher = zmesh.Mesher((1,1,1))
    mesher.mesh(labels)

    N = 1
    start = time.time()
    for label in tqdm(mesher.ids()):
        mesher.get_mesh(label, 
            simplification_factor=100, 

            # Max tolerable error in physical distance
            max_simplification_error=1,
        )
    end = time.time()
    result("simplification (connectomics.npy)", end - start, labels, N=N)

print("ZMESH")
test_zmesh_marching_cubes()
print("SKIMAGE")
test_scikit_marching_cubes()
print("ZMESH SIMPLIFICATION")
test_zmesh_simplification()


import numpy as np
import zmesh
import time

labels = np.load("connectomics.npy")
mesher = zmesh.Mesher( (32, 32, 40) )

s = time.time()
mesher.mesh(labels)
print(f"MC: {time.time() - s:.3f} sec")

s = time.time()
mesh = mesher.get_mesh(labels[400,400,128], normals=False, simplification_factor=0, max_simplification_error=40)
print(f"Simplify: {time.time() - s:.3f} sec")

with open('mesh.ply', 'bw') as f:
  f.write(mesh.to_ply())



import numpy as np
import re
import struct

class Mesh:
  """
  Represents the vertices, faces, and normals of a mesh
  as numpy arrays.

  class Mesh:
    ndarray[float32, ndim=2] self.vertices: [ [x,y,z], ... ]
    ndarray[uint32,  ndim=2] self.faces:    [ [v1,v2,v3], ... ]
    ndarray[float32, ndim=2] self.normals:  [ [nx,ny,nz], ... ]

  """
  def __init__(self, vertices, faces, normals, id=None):
    self.vertices = vertices
    self.faces = faces
    self.normals = normals
    self.id = id

  def __len__(self):
    return self.vertices.shape[0]

  def __eq__(self, other):
    """Tests strict equality between two meshes."""

    no_self_normals = self.normals is None or self.normals.size == 0
    no_other_normals = other.normals is None or other.normals.size == 0

    if no_self_normals != no_other_normals:
      return False
       
    equality = np.all(self.vertices == other.vertices) \
      and np.all(self.faces == other.faces)

    if no_self_normals:
      return equality

    return (equality and np.all(self.normals == other.normals))

  def __repr__(self):
    return "Mesh(vertices<{}>, faces<{}>, normals<{}>)".format(
      self.vertices.shape[0], self.faces.shape[0], 
      (None if self.normals is None else self.normals.shape[0])
    )

  @property
  def nbytes(self) -> int:
    nbytes = self.vertices.nbytes if self.vertices is not None else 0
    nbytes += self.faces.nbytes if self.faces is not None else 0
    nbytes += self.normals.nbytes if self.normals is not None else 0
    return nbytes

  def clone(self):
    if self.normals is None:
      return Mesh(np.copy(self.vertices), np.copy(self.faces), None)
    else:
      return Mesh(np.copy(self.vertices), np.copy(self.faces), np.copy(self.normals))

  def triangles(self):
    Nf = self.faces.shape[0]
    tris = np.zeros( (Nf, 3, 3), dtype=np.float32, order='C' ) # triangle, vertices, (x,y,z)

    for i in range(Nf):
      for j in range(3):
        tris[i,j,:] = self.vertices[ self.faces[i,j] ]

    return tris

  @classmethod
  def from_precomputed(self, binary):
    """
    Mesh from_precomputed(self, binary)

    Decode a Precomputed format mesh from a byte string.
    
    Format:
      uint32        Nv * float32 * 3   uint32 * 3 until end
      Nv            (x,y,z)            (v1,v2,v2)
      N Vertices    Vertices           Faces
    """
    num_vertices = struct.unpack("=I", binary[0:4])[0]
    try:
      # count=-1 means all data in buffer
      vertices = np.frombuffer(binary, dtype=np.float32, count=3*num_vertices, offset=4)
      faces = np.frombuffer(binary, dtype=np.uint32, count=-1, offset=(4 + 12 * num_vertices)) 
    except ValueError:
      raise ValueError("""
        The input buffer is too small for the Precomputed format.
        Minimum Bytes: {} 
        Actual Bytes: {}
      """.format(4 + 4 * num_vertices, len(binary)))

    vertices = vertices.reshape(num_vertices, 3)
    faces = faces.reshape(faces.size // 3, 3)

    return Mesh(vertices, faces, normals=None)

  def to_precomputed(self):
    """
    bytes to_precomputed(self)

    Convert mesh into binary format compatible with Neuroglancer.
    Does not preserve normals.
    """
    vertex_index_format = [
      np.uint32(self.vertices.shape[0]), # Number of vertices (3 coordinates)
      self.vertices,
      self.faces
    ]
    return b''.join([ array.tobytes('C') for array in vertex_index_format ])

  @classmethod
  def from_obj(self, text):
    """Given a string representing a Wavefront OBJ file, decode to a zmesh.Mesh."""

    vertices = []
    faces = []
    normals = []

    if type(text) is bytes:
      text = text.decode('utf8')

    for line in text.split('\n'):
      line = line.strip()
      if len(line) == 0:
        continue
      elif line[0] == '#':
        continue
      elif line[0] == 'f':
        if line.find('/') != -1:
          # e.g. f 6092/2095/6079 6087/2092/6075 6088/2097/6081
          # i.e. f vertex_1/texture_1/normal_1 etc
          (v1, vt1, vn1, v2, vt2, vn2, v3, vt3, vn3) = re.match(r'f\s+(\d+)/(\d*)?/(\d+)?\s+(\d+)/(\d*)?/(\d+)?\s+(\d+)/(\d*)?/(\d+)?', line).groups()
        else:
          (v1, v2, v3) = re.match(r'f\s+(\d+)\s+(\d+)\s+(\d+)', line).groups()
        faces.append( (int(v1), int(v2), int(v3)) )
      elif line[0] == 'v':
        if line[1] == 't': # vertex textures not supported
          # e.g. vt 0.351192 0.337058
          continue 
        elif line[1] == 'n': # vertex normals
          # e.g. vn 0.992266 -0.033290 -0.119585
          (n1, n2, n3) = re.match(r'vn\s+([-\d\.e]+)\s+([-\d\.e]+)\s+([-\d\.e]+)', line).groups()
          normals.append( (float(n1), float(n2), float(n3)) )
        else:
          # e.g. v -0.317868 -0.000526 -0.251834
          (v1, v2, v3) = re.match(r'v\s+([-\d\.e]+)\s+([-\d\.e]+)\s+([-\d\.e]+)', line).groups()
          vertices.append( (float(v1), float(v2), float(v3)) )

    vertices = np.array(vertices, dtype=np.float32)
    faces = np.array(faces, dtype=np.uint32)
    normals = np.array(normals, dtype=np.float32)

    return Mesh(vertices, faces - 1, normals)

  def to_obj(self):
    """Return a string representing a .obj file."""
    objdata = []
    objdata += [ 'v {:.5f} {:.5f} {:.5f}'.format(*vertex) for vertex in self.vertices ]
    objdata += [ 'f {} {} {}'.format(*face) for face in (self.faces+1) ] # obj is 1 indexed
    objdata = '\n'.join(objdata) + '\n'
    return objdata.encode('utf8')

  @classmethod
  def from_ply(self, plydata):
    """
    Read from a binary representation formated as ply
    Note that this code is limited to parse the ply file saved by to_ply function
    It do not support ply file written by other softwares.
    """
    header, _, data = plydata.partition(b'end_header\n')
    lines = header.splitlines()
    assert lines[0] == b'ply'
    assert lines[1] == b'format binary_little_endian 1.0'
    vertexct = int(lines[2].decode('utf-8').split()[-1])
    trianglect = int(lines[6].decode('utf-8').split()[-1])

    # interpreate the data
    vertices = np.frombuffer(data, dtype=np.float32, count=3*vertexct, offset=0).reshape(vertexct, 3)
    faces = np.frombuffer(data, dtype=np.uint32, offset=vertexct*3*4).reshape(trianglect, 4)[:, 1:]
    return Mesh(vertices, faces, normals=None) 

  def to_ply(self):
    """
    Return a bytearray in .ply format, 
    a more compact format than .obj that's still widely readable.
    """
    vertexct = self.vertices.shape[0]
    trianglect = self.faces.shape[0]

    # Header
    plydata = bytearray("""ply
format binary_little_endian 1.0
element vertex {}
property float x
property float y
property float z
element face {}
property list int int vertex_indices
end_header
""".format(vertexct, trianglect).encode('utf8'))

    # Vertex data (x y z): "fff" 
    plydata.extend(self.vertices.tobytes('C'))

    # Faces (3 f1 f2 f3): "3iii" 
    plydata.extend(
      np.insert(self.faces, 0, 3, axis=1).tobytes('C')
    )

    return plydata

  def viewer(self):
    # thanks to ChatGPT for making it easy to figure out
    # how to display VTK meshes.
    import vtk
    polydata = _create_vtk_mesh(self.vertices, self.faces)

    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(polydata)

    actor = vtk.vtkActor()
    actor.SetMapper(mapper)

    renderer = vtk.vtkRenderer()
    render_window = vtk.vtkRenderWindow()
    render_window.AddRenderer(renderer)
    render_window_interactor = vtk.vtkRenderWindowInteractor()
    render_window_interactor.SetRenderWindow(render_window)

    render_window.SetSize(1024, 1024)

    renderer.AddActor(actor)
    renderer.SetBackground(0.1, 0.2, 0.3)  # Background color

    render_window.Render()
    render_window_interactor.Start()

def _create_vtk_mesh(vertices, faces):
  import vtk
  from vtk.util.numpy_support import numpy_to_vtk, numpy_to_vtkIdTypeArray

  vtk_points = vtk.vtkPoints()
  vtk_points.SetData(numpy_to_vtk(vertices))
  
  polydata = vtk.vtkPolyData()
  polydata.SetPoints(vtk_points)
  
  vtk_faces = vtk.vtkCellArray()
  vtk_faces.SetCells(
    faces.shape[0], 
    numpy_to_vtkIdTypeArray(np.hstack([np.full((faces.shape[0], 1), 3), faces]).flatten())
  )

  polydata.SetPolys(vtk_faces)
  
  return polydata





from jinja2 import Template

with open("./templates/mesherclass.j2", "rt") as f:
	text = f.read()

template = Template(text)

for position_type in [32, 64]:
	for label_type in [8, 16, 32, 64]:
		print(template.render(
			position_type=position_type, label_type=label_type
		))


import pytest
import numpy as np
import gzip
import sys

import zmesh

DTYPE = [ np.uint8, np.uint16, np.uint32, np.uint64 ]

@pytest.fixture
def connectomics_labels():
  with gzip.open('./connectomics.npy.gz', 'rb') as f:
    return np.load(f)

@pytest.fixture
def fanc_label():
  with gzip.open('./fanc_bug.npy.gz', 'rb') as f:
    return np.load(f)[...,0]

@pytest.mark.parametrize("dtype", DTYPE)
@pytest.mark.parametrize("close", [ False, True ])
@pytest.mark.parametrize("order", [ 'C', 'F' ])
def test_executes_legacy(dtype, close, order):
  labels = np.zeros( (11,17,19), dtype=dtype, order=order)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels, close=close)

  mesh = mesher.get_mesh(1, normals=False)
  assert len(mesh.vertices) > 0
  assert len(mesh.faces) > 0
  assert mesh.normals is None

  mesh = mesher.get_mesh(1, normals=True)
  assert len(mesh.vertices) > 0
  assert len(mesh.faces) > 0
  assert len(mesh.normals) > 0


@pytest.mark.parametrize("dtype", DTYPE)
@pytest.mark.parametrize("close", [ False, True ])
@pytest.mark.parametrize("order", [ 'C', 'F' ])
def test_executes(dtype, close, order):
  labels = np.zeros( (11,17,19), dtype=dtype, order=order)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels, close=close)

  mesh = mesher.get(1, normals=False)
  assert len(mesh.vertices) > 0
  assert len(mesh.faces) > 0
  assert mesh.normals is None

  mesh = mesher.get(1, normals=True)
  assert len(mesh.vertices) > 0
  assert len(mesh.faces) > 0
  assert len(mesh.normals) > 0

@pytest.mark.parametrize("dtype", DTYPE)
@pytest.mark.parametrize("order", [ 'C', 'F' ])
def test_simplify(dtype, order):
  labels = np.zeros( (11,17,19), dtype=dtype, order=order)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels)

  mesh = mesher.get_mesh(1, normals=False)
  Nv = len(mesh.vertices)
  Nf = len(mesh.faces)
  
  mesh = mesher.simplify(mesh, reduction_factor=10, max_error=40)
  assert len(mesh) > 0
  assert len(mesh) < Nv 
  assert mesh.faces.shape[0] < Nf
  assert mesh.normals is None or mesh.normals.size == 0

  mesh = mesher.simplify(mesh, reduction_factor=10, max_error=40, compute_normals=True)
  assert len(mesh) > 0
  assert len(mesh) < Nv 
  assert mesh.faces.shape[0] < Nf
  assert mesh.normals.size > 0

  mesher.voxel_res = (1,1,1)
  mesh = mesher.get_mesh(1, normals=False)

  # ensure negative vertices work
  mesh.vertices -= 10000
  mesh = mesher.simplify(mesh, reduction_factor=2, max_error=40, compute_normals=True)
  assert len(mesh) > 0
  assert len(mesh) <= Nv 
  assert mesh.faces.shape[0] <= Nf
  assert mesh.normals.size > 0

  # check that upper limit of precision errors
  try:
    mesh.vertices[0] = 0
    mesh.vertices[1:] = 2**20
    mesh = mesher.simplify(mesh, reduction_factor=2, max_error=40, compute_normals=True)
    assert False
  except ValueError:
    pass

@pytest.mark.parametrize("dtype", DTYPE)
def test_precomputed_legacy(dtype):
  labels = np.zeros( (11,17,19), dtype=dtype)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels)
  mesh = mesher.get_mesh(1, normals=False)

  precomputed_mesh = mesh.to_precomputed()
  reconstituted = zmesh.Mesh.from_precomputed(precomputed_mesh)
  assert reconstituted == mesh

  mesh = mesher.get_mesh(1, normals=True)

  precomputed_mesh = mesh.to_precomputed()
  reconstituted = zmesh.Mesh.from_precomputed(precomputed_mesh)
  assert reconstituted != mesh # Precomputed doesn't preserve normals

@pytest.mark.parametrize("dtype", DTYPE)
def test_precomputed(dtype):
  labels = np.zeros( (11,17,19), dtype=dtype)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels)
  mesh = mesher.get(1, normals=False)

  precomputed_mesh = mesh.to_precomputed()
  reconstituted = zmesh.Mesh.from_precomputed(precomputed_mesh)
  assert reconstituted == mesh

  mesh = mesher.get(1, normals=True)

  precomputed_mesh = mesh.to_precomputed()
  reconstituted = zmesh.Mesh.from_precomputed(precomputed_mesh)
  assert reconstituted != mesh # Precomputed doesn't preserve normals

def test_obj_import():
  labels = np.zeros( (11,17,19), dtype=np.uint32)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels)
  mesh = mesher.get(1, normals=False)

  obj_str = mesh.to_obj()
  mesh2 = zmesh.Mesh.from_obj(obj_str)
  
  assert mesh == mesh2

def test_ply_import():
  labels = np.zeros( (11,17,19), dtype=np.uint32)
  labels[1:-1, 1:-1, 1:-1] = 1

  mesher = zmesh.Mesher( (4,4,40) )
  mesher.mesh(labels)
  mesh = mesher.get(1, normals=False)

  plydata = mesh.to_ply()
  mesh2 = zmesh.Mesh.from_ply(plydata)
  
  assert mesh == mesh2

def test_C_F_meshes_same_legacy(connectomics_labels):
  connectomics_labels = connectomics_labels[102:,31:,17:]

  fdata = np.asfortranarray(connectomics_labels)
  cdata = np.ascontiguousarray(connectomics_labels)

  f_mesher = zmesh.Mesher((1,1,1))
  f_mesher.mesh(fdata)

  c_mesher = zmesh.Mesher((1,1,1))
  c_mesher.mesh(cdata)

  cids = c_mesher.ids()
  cids.sort()
  fids = f_mesher.ids()
  fids.sort()
  assert cids == fids

  for label in c_mesher.ids()[:300]:
    c_mesh = c_mesher.get_mesh(label, normals=False, simplification_factor=0)
    f_mesh = f_mesher.get_mesh(label, normals=False, simplification_factor=0)
    assert np.isclose(c_mesh.vertices.mean(), f_mesh.vertices.mean())

@pytest.mark.parametrize("transpose", [True,False])
def test_fanc_bug(fanc_label, transpose):
  if transpose:
    fanc_label = fanc_label.T
  fdata = np.asfortranarray(fanc_label)
  cdata = np.ascontiguousarray(fanc_label)

  f_mesher = zmesh.Mesher((1,1,1))
  f_mesher.mesh(fdata)

  c_mesher = zmesh.Mesher((1,1,1))
  c_mesher.mesh(cdata)

  assert c_mesher.ids() == f_mesher.ids()

  for label in c_mesher.ids():
    c_mesh = c_mesher.get(label, normals=False, reduction_factor=0)
    f_mesh = f_mesher.get(label, normals=False, reduction_factor=0)
    assert np.isclose(c_mesh.vertices.mean(), f_mesh.vertices.mean())

@pytest.mark.parametrize("order", [ 'C', 'F' ])
def test_unsimplified_meshes_remain_the_same(connectomics_labels, order):
  if order == "C":
    connectomics_labels = np.ascontiguousarray(connectomics_labels)
  else:
    connectomics_labels = np.asfortranarray(connectomics_labels)

  mesher = zmesh.Mesher( (32,32,40) )
  mesher.mesh(connectomics_labels)

  for lbl in mesher.ids()[:300]:
    with gzip.open(f"./connectomics_npy_meshes/unsimplified/{lbl}.ply.gz", "rb") as f:
      old_mesh = zmesh.Mesh.from_ply(f.read())
    new_mesh = mesher.get_mesh(lbl, normals=False, simplification_factor=0, max_simplification_error=40)
    assert np.all(np.sort(old_mesh.vertices[old_mesh.faces], axis=0) == np.sort(new_mesh.vertices[new_mesh.faces], axis=0))
    print(lbl, "ok")


  mesher = zmesh.Mesher( (1,1,1) )
  mesher.mesh(connectomics_labels)

  mesher2 = zmesh.Mesher( (1,1,1) )
  mesher2.mesh(connectomics_labels)

  for lbl in mesher.ids()[:50]:
    old_mesh = mesher.get_mesh(lbl, normals=False, simplification_factor=0, max_simplification_error=40)
    new_mesh = mesher2.get(lbl, normals=False, reduction_factor=0, max_error=40)

    old_pts = old_mesh.vertices[old_mesh.faces]
    old_pts = old_pts.reshape(old_pts.shape[0] * old_pts.shape[1], 3)
    transposed_old = np.copy(old_pts)
    transposed_old[:,0] = old_pts[:,2]
    transposed_old[:,2] = old_pts[:,0]
    del old_pts
    transposed_old.sort(axis=0)

    new_pts = new_mesh.vertices[new_mesh.faces]
    new_pts = new_pts.reshape(new_pts.shape[0] * new_pts.shape[1], 3)
    new_pts.sort(axis=0)

    assert np.all(transposed_old == new_pts)
    print(lbl, "ok")  

# F order meshes are processed in a different order and so
# the simplifier produces a different mesh. Will have to add F order examples
# in order to test.
@pytest.mark.parametrize("order", [ 'C' ])
@pytest.mark.skipif(sys.platform != 'darwin', reason="Different implementations of unordered_map on different platforms have different iteration behavior. Only MacOS will match.")
def test_simplified_meshes_remain_the_same(connectomics_labels, order):
  if order == "C":
    connectomics_labels = np.ascontiguousarray(connectomics_labels)
  else:
    connectomics_labels = np.asfortranarray(connectomics_labels)

  mesher = zmesh.Mesher( (32,32,40) )
  mesher.mesh(connectomics_labels)

  for lbl in mesher.ids()[:300]:
    with gzip.open(f"./connectomics_npy_meshes/simplified/{lbl}.ply.gz", "rb") as f:
      old_mesh = zmesh.Mesh.from_ply(f.read())
    new_mesh = mesher.get_mesh(lbl, normals=False, simplification_factor=100, max_simplification_error=40)
    assert np.all(np.isclose(np.sort(old_mesh.vertices[old_mesh.faces], axis=0), np.sort(new_mesh.vertices[new_mesh.faces], axis=0)))
    print(lbl, "ok")


@pytest.mark.parametrize("reduction_factor", [2,5,10])
def test_min_error_skip(reduction_factor):
  for order in ['C','F']:
    labels = np.zeros((11, 17, 19), dtype=np.uint8, order=order)
    labels[1:-1, 1:-1, 1:-1] = 1

    mesher = zmesh.Mesher((4, 4, 40))
    mesher.mesh(labels)

    original_mesh = mesher.get_mesh(1, normals=False)
    mesh = mesher.simplify(
        original_mesh, 
        reduction_factor=reduction_factor, 
        max_error=40, 
        min_error=0,
        compute_normals=True
    )
    factor = len(original_mesh.faces) / len(mesh.faces)
    assert abs(factor - reduction_factor) < 1






#!/usr/bin/env python

import os
import setuptools
import sys

class NumpyImport:
  def __repr__(self):
    import numpy as np

    return np.get_include()

  __fspath__ = __repr__


# NOTE: If zmesh.cpp does not exist, you must run
# cython --cplus -I./zi_lib/ zmesh.pyx

extra_compile_args = []
if sys.platform == 'win32':
  extra_compile_args += [
    '/std:c++17', '/O2'
  ]
else:
  extra_compile_args += [
    '-std=c++17', '-O3', '-Wno-unused-local-typedefs', 
    '-DNDEBUG',
  ]

include_dirs = [ str(NumpyImport()), 'zi_lib/', './' ]

setuptools.setup(
  setup_requires=['pbr', 'numpy', 'cython'],
  python_requires=">=3.8",
  pbr=True,
  define_macros=[ ("NDEBUG", 1) ],
  extras_require={
    "viewer": [ "vtk" ],
  },
  ext_modules=[
    setuptools.Extension(
      'zmesh._zmesh',
      sources=[ 'zmesh/_zmesh.pyx' ],
      depends=[ 'cMesher.hpp' ],
      language='c++',
      include_dirs=include_dirs,
      extra_compile_args=extra_compile_args
    ),
  ],
)




import numpy as np
import zmesh

labels = np.zeros( (50, 50, 50), dtype=np.uint32)
labels[1:-1, 1:-1, 1:-1] = 1

mesher = zmesh.Mesher( (3.141, 1, 1) )
mesher.mesh(labels)
mesh = mesher.get_mesh(1, normals=False, simplification_factor=100, max_simplification_error=100)

print(mesh)

with open('wow.obj', 'bw') as f:
  f.write(mesh.to_obj())

with open('wow.ply', 'bw') as f:
  f.write(mesh.to_ply())

